use std::fs::{self, read_to_string};

use proc_macro::TokenStream;
use quote::{format_ident, quote};
use syn::{parse, LitBool, Result};
use syn::{
    parse::{Parse, ParseStream},
    spanned::Spanned,
    Error, Ident, Lit, LitInt,
};

const DATA_DIR: &str = "data";
const EXAMPLE_DIR: &str = "examples";
const VAR_NAME: &str = "DATA";

#[derive(Debug, Clone)]
struct IncludeData {
    var: String,
    year: Option<String>,
    day: Option<String>,
}

impl Default for IncludeData {
    fn default() -> Self {
        IncludeData {
            var: VAR_NAME.into(),
            year: None,
            day: None,
        }
    }
}

fn get_text<T: Spanned>(item: T) -> Result<String> {
    let span = item.span();
    match span.source_text() {
        Some(text) => Ok(text),
        None => {
            let start = span.start();
            let end = span.end();
            Err(Error::new(span, format!("Failed to get sourcetext for {}:{}-{}:{}", start.line, start.column, end.line, end.column,)))
        },
    }
}

impl Parse for IncludeData {
    fn parse(input: ParseStream) -> Result<Self> {
        let mut data = IncludeData::default();
        if input.peek(Ident) && !input.peek(LitInt) {
            data.var = get_text(input.parse::<Ident>().unwrap())?;
        }
        data.year = Some(get_text(input.parse::<LitInt>()?)?);
        data.day = Some(get_text(input.parse::<Lit>()?)?);

        Ok(data)
    }
}

fn canonicalize(dir: &str, input: IncludeData) -> Option<String> {
    let pathname = format!("{}/{}/{}.txt", dir, input.year.unwrap(), input.day.unwrap());
    match fs::canonicalize(pathname) {
        Ok(canon) => {
            if canon.is_file() {
                match canon.to_str() {
                    Some(c) => Some(c.to_string()),
                    None => None,
                }
            } else {
                None
            }
        },
        Err(_) => None,
    }
}

/// includes Data from Advent of code
/// ```ignore (cannot-doctest-external-file-dependency)
/// include_data!(DATA 2015 01)
/// // or
/// include_data!(2015 01)
///
/// fn main() {
///     print!("{DATA}");
/// }
/// ```
/// The Equivalent is
/// ```ignore (cannot-doctest-external-file-dependency)
/// static DATA: &str = include_str!("../../../data/2015/01.txt");
/// ```
#[proc_macro]
pub fn include_data(data: TokenStream) -> TokenStream {
    let input = match parse::<IncludeData>(data) {
        Ok(include) => include,
        Err(error) => return error.into_compile_error().into(),
    };
    let ident = format_ident!("{}", input.var);
    let mut comment: String;
    if input.day.is_some() && input.year.is_some() {
        comment = "Data from the data dir".into();
        let mut path = canonicalize(DATA_DIR, input.clone());
        if path.is_none() {
            comment = "Data from the examples".into();
            path = canonicalize(EXAMPLE_DIR, input)
        }
        match path {
            Some(p) => {
                return quote! {
                    #[doc = #comment]
                    const #ident: &str = include_str!(#p).trim_ascii();
                }
                .into();
            },
            None => comment = "failed to get data from the paths".into(),
        }
    } else {
        comment = format!("Failed to get the year({:?}) or day({:?}) falling back to default", input.year, input.day);
    }
    quote! {
        #[doc = #comment]
        const #ident: &str = "";
    }
    .into()
}

/// Same as include_data!, but it only gets data from the example directory.
/// This is useful for testing the examples, even when Data from the AOC is available
#[proc_macro]
pub fn include_example(data: TokenStream) -> TokenStream {
    let input = match parse::<IncludeData>(data) {
        Ok(include) => include,
        Err(error) => return error.into_compile_error().into(),
    };
    let ident = format_ident!("{}", input.var);
    let mut comment: String;
    if input.day.is_some() && input.year.is_some() {
        comment = "Data from the Example dir".into();
        match canonicalize(EXAMPLE_DIR, input.clone()) {
            Some(p) => {
                return quote! {
                    #[doc = #comment]
                    const #ident: &str = include_str!(#p).trim_ascii();
                }
                .into();
            },
            None => comment = "failed to get data from the path".into(),
        }
    } else {
        comment = format!("Failed to get the year({:?}) or day({:?}) falling back to default", input.year, input.day);
    }
    quote! {
        #[doc = #comment]
        const #ident: &str = "";
    }
    .into()
}

#[derive(Debug, Clone, Default)]
struct NewIncludeData {
    var: String,
    year: String,
    day: String,
    example: bool,
    parts: usize,
}

impl Parse for NewIncludeData {
    fn parse(input: ParseStream) -> Result<Self> {
        let mut v = Self::default();
        v.var = if input.peek(Ident) && !input.peek(LitInt) {
            get_text(input.parse::<Ident>().unwrap())?
        } else {
            VAR_NAME.to_string()
        };
        v.year = get_text(input.parse::<LitInt>()?)?;
        v.day = get_text(input.parse::<Lit>()?)?;
        if input.peek(LitBool) {
            v.example = input.parse::<LitBool>()?.value;
        }
        if input.peek(LitInt) {
            v.parts = input.parse::<LitInt>()?.base10_parse()?;
        }
        if v.parts == 0 {
            v.parts = 1;
        }
        Ok(v)
    }
}

fn fallback(input: NewIncludeData, message: String) -> TokenStream {
    if input.parts != 1 {
        let postfixes = (1..=input.parts).map(|i| format_ident!("{}_{}", input.var, i));
        quote! {
            #(
                ///
                const #postfixes: &str = "";
            )*
        }
        .into()
    } else {
        let name = format_ident!("{}", input.var);
        quote! {
            /// #name
            const #name: &str = "";
        }
        .into()
    }
}

/// Macro to include data and examples.
/// The First Argument(Optional) is the name/prefix for the constant,
/// The Second is the Year
/// The Third is the day/filename without the .txt at the end
/// the fourth is if its an example or not
/// and the last is the amount of parts in the file, if it is greater than 1 the first argument is treated as an Prefix.
#[proc_macro]
pub fn include_aoc(data: TokenStream) -> TokenStream {
    let input = match parse::<NewIncludeData>(data) {
        Ok(data) => data,
        Err(error) => return error.into_compile_error().into(),
    };
    let mut content = Default::default();
    if !input.example {
        let path = format!("{}/{}/{}.txt", DATA_DIR, input.year, input.day);
        match fs::exists(path.clone()) {
            Ok(true) => content = read_to_string(path).unwrap(),
            Ok(false) => {},
            Err(error) => return fallback(input, format!("Failed to read file: {}", error)),
        };
    }
    if content.is_empty() {
        let path = format!("{}/{}/{}.txt", EXAMPLE_DIR, input.year, input.day);
        content = match fs::exists(path.clone()) {
            Ok(true) => read_to_string(path).unwrap(),
            Ok(false) => Default::default(),
            Err(error) => return fallback(input, format!("Failed to read file: {}", error)),
        };
    }
    if input.parts != 1 {
        let parts = content.splitn(input.parts, "\n\n");
        let suffix = (1..=input.parts).map(|i| format_ident!("{}_{}", input.var, i));
        quote! {
            #(
                const #suffix: &str = #parts;
            )*
        }
    } else {
        let name = format_ident!("{}", input.var);
        quote! {
            const #name: &str = #content;
        }
    }
    .into()
}
